function r = write_ensblk(instclass, filelist, params, varargin)
% function write_ensblk(instclass, rawfilelist, params, ...)
%
% Writes the command (*.cmd) and log (*.log) and binary (*.bin) files
% required by ldcodas, starting from UHDAS  logging
% This file breaks the correspondence between raw file base and 
% cmd filename.  Instead, .cmd files correspond to blocks
%
% It also writes out files with (t x y) in a format suitable
% for refabs input.  The first fix from the ensemble goes to
% *.gps1, the last to *.gps2.
%
% Required arguments:
%   instclass:              instrument class (nb, bb, os, son)
%   file: name of RAW
%   params: a structure with fields
%     params.h_align      : Heading alignment in degrees;
%                             RDI sign convention
%     params.velscale      : multiply velocities by this factor
%     params.interval     : Averaging interval in seconds (default 300)
%     params.end_dday     : Averaging interval end times (as from a 
%                             codas3 database; eg. to match pingdata times)  
%                             default empty (disabled).  
%                             Overrides params.interval if set.  
%     params.ilayer       : bin0:bin1, reference layer indices
%                             for averaging.  Give the array of
%                             indices, not just the first and last bin.
%     params.geometry     : 'convex' (OS) 'concave' (NB) 
%     params.yearbase     : yearbase (required for nbuh and sonuh)
%     params.ducer_depth      : add to transducer pressure;
%                             (nbuh and sonuh start with tr_pressure = 0)
%
%     params.use_refavg   : def:1 (0, use simple reference layer average)
%     params.min_pings    : Minimum number of pings for an average;
%                             only relevant at the end of file with
%                             "finish" set to 1 (see below)
%     params.pingedit     :         level of pingediting to apply
%                                   'none' (or 0)   no pingediting at all
%                                   'ampbt'         amp and bottom only
%                                   'all'  (or 1)   amp and bottom, and water
%                                   'debug'         return both answers
%                                                   'none' and 'all' as
%                                                  a.enu_noedit, a.enul_pedit
%                                                   (a.enu is a.enu_pingedit)
%     params.min_pg       : Within the reference layer, use only
%                             the depths with at least this percent good;
%                             if there are not at least 2 such levels,
%                             don't use a reference layer, just calculate
%                             a simple average.
%                             NOTE: decrease this significantly if pingediting
%     params.use_3_beam   : Not yet implemented; only 4-beam solutions
%                             are used at present
%     params.instrument   : used only to select specific irregular transformations
%                         :        (two exist: "kmson", "kmos")
%
%   >>The first four param fields are required;<<
%   the remainder have defaults.  Read the source code (pingedit, write_ensblk.m) to 
%   see what they are, or eliminate all uncertainty by specifying all
%   param fields explicitly.
%
%   Other fields will likely be added as additional screening
%   methods are implemented.
%
% Optional named arguments: (keyword, value pairs)
%   "ldcodas_path", bindir: directory to write *.cmd, *.bin
%   "pingtype",           : (default = 'first')
%                         :  options -
%                         : 'first' (same as 'second_set' == 0) only use bb pings
%                         : 'bb'    (same as 'second_set' == 0) only use bb pings
%                         : 'second' (same as 'second_set' == 1) only use 2nd set
%                         : 'nb'    use logic to test if file is nb only or 
%                         :         interleaved.  Only use nb pings
%   "heading_offset", dh  : transducer heading offset in degrees
%     "matfile", [0 or 1] : def:0 write a matfile with the averages
%  showverbose", [0 or 1] : def:0 (1: show raw data -- timeconsuming)
%     "gbindir", path     : path to gbin dirs (default is  '../gbin')
%
%        (second_set is normally 0, the default; set it to 1 only
%        if you have interleaved BB and NB pings and want to load
%        the NB part of the data set; leave it at 0 for the BB data,
%        or for any case when only BB or NB pings were collected.)
%
%  pingediting parameters come from get_pedefaults.m and are overridden as
%  elements of a substructure 'editcfg', eg:
%      params.editcfg.rl_range = [1:10];
%
% 2000/09/22 EF
% 2001/01/25 JH adding instclass, morphing for NBADCP and gbin;
%                    get nav fields from gbin files for sonuh and nbuh
% 2004/06 JH break change correspondence (old: raw file <--> cmd file)
%                                        (new: cmd file <--> blk file)
% (also breaking the following nice touches in load_raw:
%             use with VmDAS 
%             choice of actual timeranges for ensembles

%%%% undocumented: (for testing only) 
%
% 'max_BLKprofiles'       :     number of profiles allowed per block.
%                               (make smaller to test 'newday')



global fid_cmd fid_bin

if (nargin <= 4)
   help(mfilename)
   error('too few variables')
end

fid_genlog = fopen('write_ensblk.log','a');


%set defaults for varargin structure 
cfg.redo = 0;
cfg.ldcodas_path = './';
cfg.nav_path = './';     
cfg.att_path = './';
cfg.use_for_heading = 'heading';
cfg.pingtype = []; %defaults to 'first';
cfg.verbose = 0;
cfg.gbindir = '../gbin';
cfg.path_ext = {'time',    'best';   
                'time',    'tim'};
p.use_3_beam = 0;    % 3-beam not yet implemented
p.min_pings = 10;
p.h_bias = 0; %hardwire here; deal with offset in h_align

cfg.max_BLKprofiles = 100;  % max number of profiles to use per block TESTING

cfg = fillstruct(cfg, varargin, 0);
% rename parameters to match code
ldcodas_path = cfg.ldcodas_path;

%%
if nargin < 4 | (~iscell(filelist) & ~isstruct(filelist)) | ~isstruct(params),
   help write_ensblk.m
   fprintf('At least 3 arguments are required:\n')
   fprintf('an instrument class, a filename and a structure.')
   error('too few input arguments ')
end


if (~isfield(params, 'h_align') | ...
    ~isfield(params, 'velscale') | ...    
    ~isfield(params, 'interval') | ...
    ~isfield(params, 'ilayer') | ...
    ~isfield(params, 'ducer_depth') | ...
    ~isfield(params, 'yearbase'));
   help(mfilename)
   fprintf('''params'' requires more fields.  You have only: \n');
   params
   fprintf('must have: h_align, velscale, interval,  ilayer, yearbase, ducer_depth \n')
   fclose(fid_genlog);  fid_genlog = 0;
   return
end
if isa(instclass, 'nb') & ~isfield(params, 'beamangle')
   fprintf('NB instrument must have explicit beamangle in params structure\n')
   fprintf('exiting\n')
   fclose(fid_genlog);  fid_genlog = 0;
   return
end

if (isempty(params.yearbase))
   r=[];
   fprintf('MUST specify yearbase\n')
   fclose(fid_genlog);  fid_genlog = 0;
   return
end

% Transfer input parameters, overriding defaults.
editcfg = get_pedefaults(instclass, cfg.pingtype);
if isfield(params, 'editcfg')
   editcfg = fillstruct(editcfg, params.editcfg, 0);
   params = rmfield(params, 'editcfg');
   params.editcfg = editcfg;
end


%% use params to pass variables around
%% this is just conveniece for matlab.  We could pass things in as variables instead
params.instclass = instclass;
params.pingtype = cfg.pingtype;
params.path_ext = cfg.path_ext;
params.gbindir = cfg.gbindir;



%%%%%%%%%%%%%%% deal with filelist %%%%%%%%%%%%%%%%
%% make sure filelist is a cellarray; 
%% extract rawdir and collapse to rawbases
if isstruct(filelist)
   rawfilelist = cell(length(filelist),1);
   for ifile=1:length(filelist)
      rawfilelist{ifile} = filelist(ifile).name;
   end
   filelist = rawfilelist;
end
% now parse into filebase and directory
% assume everything comes from the same directory
rawdir = fileparts(filelist{1});
rawbaselist = cell(size(filelist));
for ifile = 1:length(rawbaselist)
   [junk, rawbaselist{ifile}] = fileparts(filelist{ifile});
end


%%%%%%%%%%%%%%%%  get ensemble info %%%%%%%%%%%%%
% find last blk log file (assumes ens_blk???.log)
% NOTE ensemble numbers start with 1, profile and blk begin with 0 
% indicate new database with: last_ensnum = 0
%                             blknum = profnum = -1
%                             blkbase_str  =''
last_ensinfo = get_last_ensinfo(instclass, rawdir, rawbaselist)
if last_ensinfo.profnum >= 0
   last_ensinfo.pingtypes = ...
       get_pingtypes(instclass, rawdir, last_ensinfo.rawfilebase, params);
end
%% profnum is comes from last_ensinfo at the outset but is controlled by 
%%     write_ensblk.m, not passed around in structures
profnum = last_ensinfo.profnum;
blknum  = last_ensinfo.blknum;
blkbase_str = sprintf('ens_blk%03d', blknum);           
if blknum == -1
   newdb = 1;
else
   newdb = 0;
end
last_ensinfo = rmfield(last_ensinfo, 'profnum');
last_ensinfo = rmfield(last_ensinfo, 'blknum');
last_ensinfo = rmfield(last_ensinfo, 'blkbase_str');
fprintf(fid_genlog, '%s: getting last_ens_info\n', nowstr);
   


%--------------------------------------------------------------------------
%%% this block is repeated at the end of the "while(another_ens)" block  
%% part (a) get info to read next ensemble                               
[ensinfo, last_ensinfo, newblock] = ...
    get_next_ensinfo(rawdir, rawbaselist, last_ensinfo, params, newdb);
profnum = profnum+1;

newdb = 0;                                                                         
another_ens = length(ensinfo);                                           
if ~another_ens                                                          
   fprintf('no data remaining\n')                                        
   return                                                                
end                                                                      
%increment the blockfile counter and reset profile number if necessary

if  newblock == 1  | profnum > cfg.max_BLKprofiles               
   % reset these 
   blknum = blknum + 1;                                                  
   blkbase_str = sprintf('ens_blk%03d', blknum);           
   % make last_ensinfo match
   last_ensinfo.profnum = 0;  
   profnum = 0;
   dbinfo = setup_dbinfo(rawdir, ensinfo, params);                                  
end

%%% deal with output files %%
% (part b) slightly different from below; only opening files here        
logfile = fullfile(ldcodas_path,    [blkbase_str  '.log'])
cmdfile = fullfile(ldcodas_path,    [blkbase_str  '.cmd']);            
binfile = fullfile(ldcodas_path,    [blkbase_str  '.bin']);             
nav1file = fullfile(ldcodas_path,   [blkbase_str  '.gps1']);            
nav2file = fullfile(ldcodas_path,   [blkbase_str  '.gps2']);            
                                                                         
fid_log  = fopen(logfile,  'a'); %% these may already exist
fid_nav1 = fopen(nav1file, 'a'); %%    so open to append
fid_nav2 = fopen(nav2file, 'a'); 
fid_cmd  = fopen(cmdfile,  'a'); 
fid_bin  = fopen(binfile,  'a'); 

% have to get this the first time whether there is a new block or not
%       to write access variables
dbinfo = setup_dbinfo(rawdir, ensinfo, params);                                  
if newblock == 1
   fprintf(fid_cmd, 'binary_file: %s\n', binfile);
   fprintf(fid_cmd, 'new_block\n');
   fprintf(fid_cmd, 'dp_mask: 1\n');
   bin_float('DEPTH', round(dbinfo.depth));
   bin_double('CONFIGURATION_1', dbinfo.config1);
   fprintf(fid_log, '%% rawfilebase    last ensemble   secs(ens)   profile_num\n');
   fprintf(fid_log, '%% -----------   -------------    -------    ------------\n');
   newblock = 0;
end

if fid_log  == -1, error(['Can''t open ', logfile]);   return, end          
if fid_nav1 == -1, error(['Can''t open ', nav1file]);  return, end        
if fid_nav2 == -1, error(['Can''t open ', nav2file]);  return, end        
if fid_cmd  == -1, error(['Can''t open ', cmdfile]);   return, end        
if fid_bin  == -1, error(['Can''t open ', binfile]);   return, end        
%--------------------------------------------------------------------------

%% do it:
while (another_ens)
   tic

   % write to block log file
  secs_in_ens = 0;
  for segi = 1:length(ensinfo)
     secs_in_ens = secs_in_ens + ensinfo(segi).secs_read;
  end
  
  if secs_in_ens > .3*params.interval    %%% set this up top
     
     fprintf(fid_log, '%20s   %10d  %10d %5d\n', ...
             ensinfo(end).rawfilebase, ...
             ensinfo(end).lastens, ...
             secs_in_ens,...
             profnum);
     fprintf(fid_genlog, '%s: (%s), %20s   %10d  %10d %5d\n', ...
             nowstr, logfile,...
             ensinfo(end).rawfilebase, ...
             ensinfo(end).lastens, ...
             secs_in_ens,...
             profnum);
     fprintf( '%s: %s  :  %15s   %8d  %3d %5d\n', ...
             nowstr, logfile,...
             ensinfo(end).rawfilebase, ...
             ensinfo(end).lastens, ...
             secs_in_ens,...
             profnum);


     %% get the data here; call avg; return raw and averaged data
     [a, r] = mk_rawens(rawdir, ensinfo, params);
     
     % all the structures:
     ancil1 = mk_ancil1(r, 1);
     ancil2 = mk_ancil2(r, 1);
     
     %NAVIGATION structure wants [latitude longitude]
     %but nav structure has [time longitude latitude]
     nav = ones(4,1) * 1e38;   

     fprintf(fid_cmd, ['new_profile: ', ...
                       to_date(a.config.yearbase, a.lastdday),...
                       '  /* %d %d */\n'], r.ens_num([1 end]));
     fprintf(fid_cmd, 'depth_range: %d %d\n', round(a.depth([1 end])));
     bin_double('ACCESS_VARIABLES', dbinfo.access);
     bin_double('ANCILLARY_1', ancil1);
     bin_double('ANCILLARY_2', ancil2);
     bin_double('BOTTOM_TRACK', a.bt);
     bin_double('NAVIGATION', nav);
     bin_float('U', a.enu(:, 1));
     bin_float('V', a.enu(:, 2));
     bin_float('W', a.enu(:, 3));
     bin_float('ERROR_VEL', a.enu(:, 4));
     bin_ubyte('AMP_SOUND_SCAT', round(mean(a.amp, 2)));
     bin_ubyte('RAW_AMP', a.amp);
     bin_ubyte('PERCENT_GOOD', a.pg(:, 4));
     bin_ubyte('PERCENT_3_BEAM', a.pg(:, 3));

     %% Maybe we should call it something else, but this will do
     %% for now; it is not correlation, either.
     if (~isfield(a, 'sw'))
        if (isfield(a, 'cor'))
           a.sw = a.cor;
        else
           a.sw = zeros(size(a.amp));
        end
     end
     bin_ubyte('SPECTRAL_WIDTH', round(mean(a.sw, 2)));
     bin_ubyte('RAW_SPECTRAL_WIDTH', a.sw);
     bin_ubyte('PROFILE_FLAGS', zeros(1, length(a.depth)));
     % .....
     
     %make sure there are values for navigation 
     %txy1 is the first good fix of the ensemble
     posi = min(find(~isnan(sum(r.nav.txy1))));
     if (~isempty(posi))
        fprintf(fid_nav1, '%12.7f %12.7f %12.7f\n', r.nav.txy1(1, posi),...
                r.nav.txy1(2, posi), r.nav.txy1(3, posi));
     end
     %txy2 is the last good fix of the ensemble
     posi = max(find(~isnan(sum(r.nav.txy2))));
     if (~isempty(posi))
        fprintf(fid_nav2, '%12.7f %12.7f %12.7f\n', r.nav.txy1(1, posi),...
                r.nav.txy1(2, posi), r.nav.txy1(3, posi));
     end

     elapsed_time = toc;
     fprintf('time used: %4.2fs\n', elapsed_time)
   
  end
  
        
  %-----------------------------------------------------------
  % set the stage for the next ensemble

  [ensinfo, last_ensinfo, newblock] = ...
      get_next_ensinfo(rawdir, rawbaselist, last_ensinfo, params, newdb);
  
  if length(ensinfo) == 0
     fprintf(' no more data. quitting here \n')
     break
  end
  
  %% evidently we still have data
  profnum = profnum+1;
  
  
  
  if  newblock == 1  | profnum >= cfg.max_BLKprofiles               
     fprintf('----------------starting new block.  closing files for blk%03d\n', blknum)
     % reset these 
     blknum = blknum + 1;                                               
     fprintf('----------------... opening files for blk%03d\n', blknum)
     blkbase_str = sprintf('ens_blk%03d', blknum);           
     % make last_ensinfo match
     profnum = 0;  
     newblock = 0;
     
     %% close the old files
     if exist('fid_bin'),  if fid_bin,  fclose(fid_bin);  fid_bin  = 0; end, end 
     if exist('fid_nav1'), if fid_nav1, fclose(fid_nav1); fid_nav1 = 0; end, end 
     if exist('fid_nav2'), if fid_nav2, fclose(fid_nav2); fid_nav2 = 0; end, end 
     if exist('fid_cmd') , if fid_cmd,  fclose(fid_cmd);  fid_cmd  = 0; end, end 
     if exist('fid_log'),  if fid_log,  fclose(fid_log);  fid_log  = 0; end, end 
                                                                                  
     % open the new files                                                    
     logfile = fullfile(ldcodas_path,    [blkbase_str  '.log'])
     cmdfile = fullfile(ldcodas_path,    [blkbase_str  '.cmd']);             
     binfile = fullfile(ldcodas_path,    [blkbase_str  '.bin']);             
     nav1file = fullfile(ldcodas_path,   [blkbase_str  '.gps1']);            
     nav2file = fullfile(ldcodas_path,   [blkbase_str  '.gps2']);            
                                                                              
     fid_log  = fopen(logfile,  'w'); % should be OK to open 'w' since 
     fid_nav1 = fopen(nav1file, 'w'); %    we're in a loop    
     fid_nav2 = fopen(nav2file, 'w'); 
     fid_cmd  = fopen(cmdfile,  'w'); 
     fid_bin  = fopen(binfile,  'w'); 
                                                                               
     %% write the new block info to the cmd file
     dbinfo = setup_dbinfo(rawdir, ensinfo, params);                                  
     fprintf(fid_cmd, 'binary_file: %s\n', binfile);
     fprintf(fid_cmd, 'new_block\n');
     fprintf(fid_cmd, 'dp_mask: 1\n');
     bin_float('DEPTH', round(dbinfo.depth));
     bin_double('CONFIGURATION_1', dbinfo.config1);
     %% write the new block info to the log file
     fprintf(fid_log, '% rawfilebase    last ensemble   secs(ens)   profile_num\n');
     fprintf(fid_log, '% -----------   -------------    -------    ------------\n');
     %% notify user
     fprintf( '------------   New Block\n');
     
     if fid_log  == -1, error(['Can''t open ', logfile]); return, end             
     if fid_nav1 == -1, error(['Can''t open ', nav1file]); return, end           
     if fid_nav2 == -1, error(['Can''t open ', nav2file]); return, end           
     if fid_cmd  == -1, error(['Can''t open ', cmdfile]);   return, end           
     if fid_bin  == -1, error(['Can''t open ', binfile]);   return, end           
     
  elseif  newblock == -1
     fprintf('ending write_blk session\n')
     break
  end                                                                            
  
  %-----------------------------------------------------------------------------------
   
end % of ensembles

% close the files
if exist('fid_bin'),  if fid_bin,  fclose(fid_bin);  fid_bin  = 0; end, end
if exist('fid_nav1'), if fid_nav1, fclose(fid_nav1); fid_nav1 = 0; end, end
if exist('fid_nav2'), if fid_nav2, fclose(fid_nav2); fid_nav2 = 0; end, end
if exist('fid_cmd') , if fid_cmd,  fclose(fid_cmd);  fid_cmd  = 0; end, end
if exist('fid_log'),  if fid_log,  fclose(fid_log);  fid_log  = 0; end, end

if exist('fid_genlog'),  if fid_genlog,  fclose(fid_genlog);  fid_genlog  = 0; end, end
